/*

Exp.java
  -- classes implementing interval arithmetic expressions,
     as part of the "ia_parser library" version 0.1beta1, 10/97

Copyright (C) 1997 Timothy J. Hickey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

 */



    /*
    ****************************************************************
    ia_parser.Exp

    ****************************************************************
    */

package ia_parser;

import java.util.*;
import ia_math.*;

/**
 * This class represents interval arithmetic expressions.
 * <p>
 * It has methods for binding the variables in the expression
 * to intervals, evaluating the expression, narrowing the
 * the variables which appear in the expression, and
 * retrieving the bindings of the variables in the expression.
 * <p>
 * These expressions consists of variables, intervals, and
 * constants combined with various function and predicate
 * symbols. Each variable in an expression has an associated
 * interval value which can be used to evaluate the expression
 * or to narrow the expression (assuming the top level operator
 * is a relation operator).
 * <p>
 * Expressions are constructed using the Parser class.
 */
public abstract class Exp {
  /**
   * Return the current interval value of the expression
   */
   public abstract RealInterval ival();

  /** 
   * Bind the variables in the expression to intervals
   * as specified by the table T. If a variable in the
   * expression does not appear in T, then it will be
   * initialized to (-infinity,infinity) and will be
   * stored in T.
   * <p>
   * If T is null then this does nothing. If T is empty,
   * but not null, then it reinitializes all variables
   * and stores them in T.
   */
   public abstract void bindVars(RealIntervalTable T);

  /**
   * Traverse the expression and store all the variable
   * bindings in the tree T. If T is initially empty
   * this will find all variables in the expression and
   * store them with their bindings in T. If T is non-empty
   * it will simply add the bindings to T. If T is null
   * it will do nothing.
   */
   public abstract void getVarBindings(RealIntervalTable T);

  /**
   * Evaluate the expression using the Table T for the values
   * of the variables. If T is null, then just use the
   * current values of the variables.
   */
   public abstract RealInterval eval(RealIntervalTable T)
      throws IAException;

  /** 
   * Narrow the variables appearing in the expression. 
   */
   public abstract boolean narrow();

}


class Num extends Exp {
  public RealInterval ival;
  public RealInterval myval;

  Num(int r) { 
    //      Debugging.printTrace(Debugging.parseTrace," Num "+r);
     myval=new RealInterval((double) r);
     ival=new RealInterval((double) r);
  }
  Num(double r) { 
    //      Debugging.printTrace(Debugging.parseTrace," Num "+r);
     myval=new RealInterval(r);
     ival=new RealInterval(r);
  }
  Num(double a, double b) {
    //      Debugging.printTrace(Debugging.parseTrace," Num= ["+a+","+b+"]");
     myval=new RealInterval(a,b); 
     ival=new RealInterval(a,b);
 }

  public String toString() {
    //    return ival.toString();
    if (ival.lo()==ival.hi())
        return(""+ival.lo());
    else
        return "["+ival.lo()+","+ival.hi()+"]";
  }

  public RealInterval ival() {return ival;}

  public void bindVars(RealIntervalTable T) { 
    //      Debugging.printTrace(Debugging.bindVarsTrace," Num "+ival.toString());
      ival=(RealInterval) myval.clone();
      return; }

  public void getVarBindings(RealIntervalTable T) {
     return;
  }

  public RealInterval eval(RealIntervalTable T)
      throws IAException
 {
    //    Debugging.printTrace(Debugging.evalTrace,
    //    "  Entering Eval Num with (ival,T) = (" + ival +" , "+ T+" ) ");
    ival=(RealInterval) myval.clone();
    return ival;
  }

  public boolean narrow() {
    //        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //        "narrowing num "+ival.toString());
    ival=(RealInterval) myval.clone();
    return true;
  }

}


class Var extends Exp {
  private String ss;

  public RealInterval ival=
      new RealInterval(Double.NEGATIVE_INFINITY,
                       Double.POSITIVE_INFINITY);

  Var(String s) { 
    ss=s;
    //    Debugging.printTrace(Debugging.parseTrace," Var "+ ss);
  }

  Var(String s,RealIntervalTable T) 
  { ss=s; 
  //    Debugging.printTrace(Debugging.parseTrace," Var "+ ss);
  //    ival = new RealInterval(Double.NEGATIVE_INFINITY,Double.POSITIVE_INFINITY);
    this.storeval(ival,T);
  }

  public String toString() {
    return ss;
    //    return ("var(" +ss+ "," +ival.toString()+ ")");
  }

  public RealInterval ival() {return ival;}

  /**
   * this uses the environment <b>T</b> to reset
   * the value of the variable
   */
  public void bindVars(RealIntervalTable T) { 
    this.eval(T);
    //                    Debugging.printTrace(Debugging.bindVarsTrace,
    //                    "bindVar "+ss+" in "+T.toString());
    return; }

  public void getVarBindings(RealIntervalTable T) {
    if (T==null) return;
    T.store(ss,ival);
  }

  public RealInterval eval(RealIntervalTable T)
      throws IAException
 {
     if (T==null) return ival;

     ival = T.lookup(ss);
     if (ival == null) {
         ival = new RealInterval(
                   Double.NEGATIVE_INFINITY,
                   Double.POSITIVE_INFINITY);
         T.store(ss,ival);
     }
     //                       Debugging.printTrace(Debugging.evalTrace,
     //                       " Eval Var = " + ss +
     //                       " ival = " + ival + " T = " + T);
     return ival;
  }

  void storeval(RealInterval x, RealIntervalTable T) {
    ival = x;
    T.store(ss,x);
  }

  public boolean narrow() {
    //        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //        "narrowing var "+ss+" = "+ival.toString());

    return true; // it would be better to check that ival is nonempty
  }

  public String Name() {return ss;}

}


class UnaryOp extends Exp {
  private Exp e1;
  private int op;
  private String s=null;

  public RealInterval ival=
      new RealInterval(Double.NEGATIVE_INFINITY,
                       Double.POSITIVE_INFINITY);


  UnaryOp(int index,Exp a1) {
    //    Debugging.printTrace(Debugging.parseTrace," Unary op "+ index);
    op=index; e1=a1;}

  UnaryOp(int index,String ss,Exp a1) {
    //    Debugging.printTrace(Debugging.parseTrace," Unary op "+ index);
    op=index; e1=a1;s=ss;}

  public String toString() {
    if (s==null)
      return ("op_"+op+"("+e1+")");
    else
      return (s+"("+e1+")");

  }

  public RealInterval ival() {return ival;}

  public void bindVars(RealIntervalTable T) { 
    //                      Debugging.printTrace(Debugging.bindVarsTrace,
    //                      " Unary op "+ op);
    ival =
      new RealInterval(Double.NEGATIVE_INFINITY,
                       Double.POSITIVE_INFINITY);
    e1.bindVars(T); 
    return; }

  public void getVarBindings(RealIntervalTable T) {
    if (T==null) return;
    e1.getVarBindings(T);
  }

  public RealInterval eval(RealIntervalTable T) 
    throws IAException
  {
    RealInterval xx;
    //    Debugging.printTrace(Debugging.evalTrace,
    //     "  Entering UnaryOp.Eval with (op,e1,T) = (" + op+" , "+e1+" , "+ T+")");
    xx = RealIntervalOp.Op(op,e1,T);
    //    Debugging.printTrace(Debugging.evalTrace,
    //      "--> "+xx);
    ival = xx;
    return xx;
  }

  public boolean narrow() {
    //        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //        "narrowing unary "+s+"("+e1.ival().toString()+") ="+
    //         ival.toString());
    return
       RealIntervalOp.narrow(op,ival,e1.ival())
       &&
       e1.narrow() 
       &&
       RealIntervalOp.narrow(op,ival,e1.ival());
  }


}


class BinaryOp extends Exp {
  protected Exp e1,e2;
  protected int op;
  protected String s=null;

  public RealInterval ival=
      new RealInterval(Double.NEGATIVE_INFINITY,
                       Double.POSITIVE_INFINITY);

  BinaryOp(int index,Exp a1,Exp a2) {
    //                        Debugging.printTrace(Debugging.parseTrace,
    //                        " Binary op "+ index);
    op=index; e1=a1; e2 = a2;}

  BinaryOp(int index,String ss,Exp a1,Exp a2) {
    //                        Debugging.printTrace(Debugging.parseTrace,
    //                        " Binary op "+ index);
    op=index; e1=a1; e2 = a2; s=ss;}


  public String toString() {
    if (s==null)
      return ("op_"+op+"("+e1+","+e2+")");
    else if (s == ";")
      return (e1+";\n"+e2);
    else
      return ("("+e1+s+e2+")");
  }

  public RealInterval ival() {return ival;}

  public void bindVars(RealIntervalTable T) { 
    //                    Debugging.printTrace(Debugging.bindVarsTrace,
    //                    " Binary op "+ op);
    ival =
      new RealInterval(Double.NEGATIVE_INFINITY,
                       Double.POSITIVE_INFINITY);
    e1.bindVars(T); 
    e2.bindVars(T); 
    return; }

  public void getVarBindings(RealIntervalTable T) {
    if (T==null) return;
    e1.getVarBindings(T);
    e2.getVarBindings(T);
  }


  public RealInterval eval(RealIntervalTable T) 
    throws IAException
  {
    //                    Debugging.printTrace(Debugging.evalTrace,
    //                    "   Entering BinaryOp.Eval with op = " + op);
    ival = RealIntervalOp.Op(op,e1,e2,T);
    //                    Debugging.printTrace(Debugging.evalTrace,
    //                    "--> "+ival);
    return ival;
  }

  public boolean narrow() {
    boolean result = false;
    //                        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //                       "narrowing binary op "+s+" index("+s+"): "
    //                        +ival.toString());
    //    System.out.println( "narrowing binary op "+s+"("+
    //                            ival+","+e1.ival()+","+e2.ival()+")");

    result =
       RealIntervalOp.narrow(op,ival,e1.ival(),e2.ival())
       &&
       e1.narrow() 
       &&
       e2.narrow()
       &&
       RealIntervalOp.narrow(op,ival,e1.ival(),e2.ival());

    //    if (result)
    //    System.out.println( "  narrowed binary op "+s+"("+
    //                            ival+","+e1.ival()+","+e2.ival()+")");
    //    else
    //    System.out.println( "  narrowing failed!");

    return result;

  }
}




/**
 *  This handles narrowing of the assignment operator Var := Exp.
 */
class ColonEqualsOp extends BinaryOp {
  private RealIntervalTable T = null;

  ColonEqualsOp(int index,Exp a1,Exp a2) {
    super(index,a1,a2);
  }
  ColonEqualsOp(int index,String ss,Exp a1,Exp a2) {
    super(index,ss,a1,a2);
  }

  public boolean narrow() {
    boolean b;

    try {
        b = RealIntervalOp.narrow(sym.COLON_EQUALS,
                  this.ival,
                  e1.ival(),
                  e2.eval(T));
        return b;
    } catch (IAException e) {
        return false;
    }
  }
}


class RealIntervalOp {
  static RealInterval Op(int index, Exp a, RealIntervalTable T) 
    throws IAException
 {
    //                         Debugging.printTrace(Debugging.evalTrace,
   //    System.out.println(
   //                             "  Entering RealIntervalOp1.Op with (index,a,T) = ("
   //                         + index+" , "+a+" ,"+ T+" ) ");
    switch (index) {
    case sym.UMINUS: return IAMath.uminus(a.eval(T));
    case sym.EXP: return IAMath.exp(a.eval(T));
    case sym.LOG: return IAMath.log(a.eval(T));
    case sym.SIN: return IAMath.sin(a.eval(T));
    case sym.COS: return IAMath.cos(a.eval(T));
    case sym.TAN: return IAMath.tan(a.eval(T));
    case sym.ASIN: return IAMath.asin(a.eval(T));
    case sym.ACOS: return IAMath.acos(a.eval(T));
    case sym.ATAN: return IAMath.atan(a.eval(T));
    case sym.MP: return IAMath.midpoint(a.eval(T));
    case sym.LP: return IAMath.leftendpoint(a.eval(T));
    case sym.RP: return IAMath.rightendpoint(a.eval(T));
    default: 
             System.out.println(
              "Error -- unknown operator with index " + index); 
             return(new RealInterval(0.0));
    }
  }

  static RealInterval Op(int index, Exp a, Exp b, RealIntervalTable T)
    throws IAException
 {
    RealInterval temp;
    //    Debugging.printTrace(Debugging.evalTrace,
    //                         "  Entering RealIntervalOp2.Op with (index,a,b,T) = ("
    //                           + index+" , "+a+" ,"+b+" ,"+ T+" ) ");
    switch (index) {
    case sym.NARROW: 
                     return new RealInterval(1.0);
    case sym.PLUS:   return IAMath.add(a.eval(T),b.eval(T));
    case sym.MINUS:  return IAMath.sub(a.eval(T),b.eval(T));
    case sym.TIMES:  return IAMath.mul(a.eval(T),b.eval(T));
    case sym.DIVIDE: return IAMath.div(a.eval(T),b.eval(T));
    case sym.CAROT:  return IAMath.integerPower(a.eval(T),b.eval(T));
    case sym.POWER:  return IAMath.power(a.eval(T),b.eval(T));
    case sym.SEMI:   temp = a.eval(T);
                     System.out.println("temp value = " + temp.toString());
                     return(b.eval(T));
    case sym.COLON_EQUALS: 
      //                     System.out.println("**** := " + a + b.eval(T));
                     ((Var) a).storeval(b.eval(T),T);return(a.eval(T));
    case sym.EQUALS: temp = a.eval(T);
                     System.out.println("constraint: " 
                        + temp.toString() + b.eval(T).toString());
                     return(temp);
    case sym.LT:
    case sym.LE:
    case sym.GT:
    case sym.GE:
    case sym.NE: temp = a.eval(T);
                     System.out.println("constraint" );return(temp);
    default: 
             System.out.println(
              "Error -- unknown operator with index " + index); 
             return(new RealInterval(0.0));
    }
  }

  // b = op(a)
  static boolean narrow(int index, RealInterval b, RealInterval a) {
    //        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //            " ** narrowing unary op "
    //             +index+" "+a.toString()+" "+b.toString());

    switch (index) {
    case sym.UMINUS:  return IANarrow.narrow_uminus(a,b);
    case sym.EXP:  return IANarrow.narrow_exp(a,b);
    case sym.LOG:  return IANarrow.narrow_log(a,b);
    case sym.SIN:  return IANarrow.narrow_sin(a,b);
    case sym.COS:  return IANarrow.narrow_cos(a,b);
    case sym.TAN:  return IANarrow.narrow_tan(a,b);
    case sym.ASIN:  return IANarrow.narrow_asin(a,b);
    case sym.ACOS:  return IANarrow.narrow_acos(a,b);
    case sym.ATAN:  return IANarrow.narrow_atan(a,b);
    case sym.MP:  return IANarrow.narrow_equals(a,b);
    case sym.LP:  return IANarrow.narrow_equals(a,b);
    case sym.RP:  return IANarrow.narrow_equals(a,b);
    default: 
             System.out.println(
              "Error -- unknown operator with index " + index); 
             return(false);
    }

  }

  /* a = b op c */
  static boolean narrow(int index, 
                        RealInterval a, RealInterval b, RealInterval c) {
    //        ExpDebugging.printTrace(ExpDebugging.narrowTrace,
    //    System.out.println(
    //            " ** narrowing binary op "
    //             +index+" "+a.toString()+" "+b.toString()+" "+c.toString()
    //             );

    switch (index) {
    case sym.NARROW: return true;
    case sym.PLUS:   return IANarrow.narrow_add(a,b,c);
    case sym.MINUS:  return IANarrow.narrow_sub(a,b,c);
    case sym.TIMES:  return IANarrow.narrow_mul(a,b,c);
    case sym.DIVIDE: return IANarrow.narrow_div(a,b,c);
    case sym.CAROT:  return IANarrow.narrow_carot(a,b,c);
    case sym.POWER:  return IANarrow.narrow_power(a,b,c);
    case sym.SEMI:   return IANarrow.narrow_semi(a,b,c);
    case sym.COLON_EQUALS: 
                 return IANarrow.narrow_colon_equals(a,b,c);
    case sym.EQUALS: 
                 return IANarrow.narrow_eq(a,b,c);
    case sym.LT: return IANarrow.narrow_lt(a,b,c);
    case sym.LE: return IANarrow.narrow_le(a,b,c);
    case sym.GT: return IANarrow.narrow_lt(a,c,b);
    case sym.GE: return IANarrow.narrow_le(a,c,b);
    case sym.NE: return IANarrow.narrow_ne(a,b,c);
    default: 
             System.out.println(
              "Error -- unknown operator with index " + index); 
             return(false);
    }
  }

}


class ExpDebugging {
  public static final int evalTrace=0;
  public static final int narrowTrace=1;
  public static final int parseTrace=2;
  public static final int bindVarsTrace=3;

   final static void printTrace(int t, String s) {
              System.out.println(t+": "+s);
  }

}
